/*******************************************************************************
 *  Licensed Materials - Property of IBM
 * (C) Copyright IBM Corp.  2001, 2006.  All Rights Reserved.
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.ccl.mapping.xsd;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.eclipse.emf.common.notify.impl.AdapterImpl;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.util.ExtendedMetaData;
import org.eclipse.xsd.XSDSchema;

import com.ibm.ccl.mapping.Mapping;
import com.ibm.ccl.mapping.MappingDesignator;
import com.ibm.ccl.mapping.MappingFactory;
import com.ibm.ccl.mapping.resolvers.IPathResolver;
import com.ibm.ccl.mapping.util.ModelUtils;
import com.ibm.ccl.mapping.xsd.XSDEcoreBuilder;
import com.ibm.ccl.mapping.xsd.resources.MappingResources;
import com.ibm.ccl.mapping.xsd.util.XSDEcoreUtils;

/**
 * XSD2EcorePathResolverAdapter implements com.ibm.ccl.mapping.resolvers.IPathResolver
 * interface. This path resolver resolves paths to and from ecore model objects representing
 * XSD model objects. Note this resolver supports incremental loading of the ecore
 * model based on access to global element, attribute and type definitions found in the 
 * underlying schema. This resolver also supports both verbose and streamline serialization
 * of the mapping resource.
 *  
 *  Syntax of path is a subset of XPATH.
 *  Path              ::=    RelativePath | AbsolutePath  
 *  AbsolutePath      ::=    VariableReference ["/" RelativePath]?
 *  RelativePath      ::=    '.' |
 *  												 AttributeNode | 
 *                           TypeNode ["/" ElementNode ]* ["/" AttributeNode ]? |
 *                           ElementNode ["/" ElementNode ]* ["/" AttributeNode ]?
 *  PathSeparator     ::=    "/"
 *  VariableReference ::=    "$"QName
 *  AttributeNode     ::=    "@"QName
 *  TypeNode          ::=    "type('"QName"')"
 *  ElementNode       ::=    QName Predicate?
 *  Predicate         ::=    "[" Digits "]"  
 *  Digits            ::=    [0-9]+ 
 *  
 *  Examples
 *  1. $in/type('PersonType')  // absolute path accessing a global type definition named personType.
 *  2. type('PersonType')      // relative path accessing type definition named personType.
 *  3. $cdlib/CD               // absolute path accessing a global element definition named CD.
 *  4. @cdlib/CD/@artist       // absolute path accessing an attribute definition.
 *  5. CD/@artist              // relative path accessing an attribute definition.
 *  
 *  This class implements the following APIs from com.ibm.ccl.mapping.resolvers.IPathResolver:
 *   public String getPathForDesignator(com.ibm.ccl.mapping.MappingDesignator, boolean)
 *   public org.eclipse.emf.ecore.EObject getModelObjectForPath(com.ibm.ccl.mapping.MappingDesignator, com.ibm.ccl.mapping.MappingDesignator, java.lang.String, org.eclipse.emf.ecore.resource.Resource)
 *   public String generateUniqueVariableName(org.eclipse.emf.ecore.EObject, java.util.Set, int)
 *   public String serializeDesignatorVariable(com.ibm.ccl.mapping.MappingDesignator)
 *   public String serializePathWithDesignatorVariable(com.ibm.ccl.mapping.MappingDesignator) 
 *   
 * @author David Lauzon lauzond@ca.ibm.com
 */

public class XSDEcorePathResolverAdapter extends AdapterImpl implements IPathResolver	
{
  // path segment constants
  public static final String PATH_SEPARATOR      = "/";       //$NON-NLS-1$
  public static final String VARIABLE_PREFIX     = "$";       //$NON-NLS-1$
  public static final String ATTRIBUTE_PREFIX    = "@";       //$NON-NLS-1$
  public static final String TYPE_FUNCTION_NAME  = "type";    //$NON-NLS-1$
  public static final String TEXT_FUNCTION_NAME  = "text";    //$NON-NLS-1$
  public static final String TEXT_NODE_NAME      = "value";   //$NON-NLS-1$
  public static final String MIXED_NAME          = "mixed";   //$NON-NLS-1$
  public static final String SingleQuote         = "\'";      //$NON-NLS-1$
  public static final String LParen              = "(";       //$NON-NLS-1$
  public static final String RParen              = ")";       //$NON-NLS-1$
  public static final String RSquareBracket      = "]";       //$NON-NLS-1$
  public static final String LSquareBracket      = "[";       //$NON-NLS-1$
  public final static String Dot                 = ".";       //$NON-NLS-1$
  public static final String DEFAULT_INPUT_VARIABLE =  "in";  //$NON-NLS-1$
  public static final String DEFAULT_OUTPUT_VARIABLE = "out"; //$NON-NLS-1$

  public final static String MAPPING_URI = "MAPPING_URI";     //$NON-NLS-1$
  
  /**
   * Access to the ecore builder used for incremental loading.
   */
  protected XSDEcoreBuilder fEcoreBuilder = new XSDEcoreBuilder();
  
  /**
   * Access to resource set for loading input and output schemas.
   * Note this resource set is the same one used for loading the mapping resource itself.
   */
  protected ResourceSet fResourceSet;

  /**
   * Access to the MappingFactory instance.
   */
  protected MappingFactory fMappingFactory = MappingFactory.eINSTANCE;

  
  /* (non-Javadoc)
   * @see com.ibm.ccl.mapping.resolvers.IPathResolver#getPath(com.ibm.ccl.mapping.MappingDesignator, java.lang.String)
   */
  public String getPath(MappingDesignator designator, String parentVariable)
  {
	String result = null;
	if (designator != null)
	{
  	  EObject modelObject = designator.getObject();
  	  if (modelObject instanceof EPackage)
  	    result = getFilePathForDesignator(designator);
  	  else if (modelObject != null)
  	  {
   	    result = getNodePathForDesignator(designator);
	    if ((parentVariable != null) && (parentVariable.length()>0))
	    {
	      result = VARIABLE_PREFIX + parentVariable + PATH_SEPARATOR + result;
	    }
  	  }
  	  else
  	  {
		result = designator.getRefName();
  	    if (designator.getIndex() != null && designator.getIndex().length() > 0)
	    {
	      result += LSquareBracket + designator.getIndex() + RSquareBracket;
	    }
	  }
	}
	return result;
  }

  
  /* (non-Javadoc)
   * @see com.ibm.ccl.mapping.resolvers.IPathResolver#resolve(com.ibm.ccl.mapping.MappingDesignator, com.ibm.ccl.mapping.MappingDesignator)
   */
  public EObject resolve(MappingDesignator designator, MappingDesignator parentDesignator)
  {
  	EObject result = null;
  	if (designator != null)
  	{
  	  String path = designator.getRefName();
      if ((path!=null) && (path.length()>0))
      {
      	Resource mappingResource =	designator.eResource();

      	// Note this assumes one schema per resource.
        if (parentDesignator==null)
        {
          // root designator -- load the file
          Resource res = loadResource(path, mappingResource);
          EPackage ePackage = getEPackageFromResource(res);
          designator.setObject(ePackage);
          result = ePackage;
        }
        else
        {
          // non-root designator
          String[] segments = path.split(PATH_SEPARATOR);
          String indexForParent = null;
          int numberOfSegments = segments.length;
          segments = removeVariableFromPathSegment(segments);
          if (segments.length < numberOfSegments)
          {
        	indexForParent = getIndex(segments[0]);
        	numberOfSegments = segments.length;
          }
          segments = removeDotFromPathSegment(segments);
          if (segments.length < numberOfSegments)
          {
       	    if (indexForParent != null && indexForParent.length() > 0)
        	  indexForParent = getIndex(segments[0]);
          }

          result = getModelObjectForPath(parentDesignator.getObject(), segments, designator, parentDesignator);
		  if (indexForParent != null && indexForParent.length() > 0)
		  {
			// need to set the index on parent's clone
			// clone is in place of first child under the parent
			MappingDesignator d = designator;
			MappingDesignator child = null;
			while (d.getParent() != parentDesignator)
			{
			  child = d;
			  d = d.getParent();
			}
			if (d.getIsParentDelta().booleanValue()) 
			{
			  // already a clone, just set index
			  d.setIndex(indexForParent);
			}
			else
			{
			  // insert a clone
			  MappingDesignator clone = ModelUtils.clone(d);
			  clone.setIsParentDelta(new Boolean(true));
			  clone.setParent(parentDesignator);
			  if (child != null)
			    child.setParent(clone);
			  // set index on the clone
			  clone.setIndex(indexForParent);
			}
		  }
        }
      }
    }
    if ((designator != null) && (designator.getParent() == null))
    {
      // a lost designator with no parent
      designator.setParent(parentDesignator);
    }
    return result;
  }
 
  /**
   * Generate a relative file path path based on the mapping 
   * resource for the specified designator.
   */
  protected String getFilePathForDesignator(MappingDesignator designator)
	{
  	String result = new String();
  	URI mappingURI = null;
    URI xsdURI = null;
    
    try
    {
  	  if (designator.eResource()!=null)
        mappingURI = designator.eResource().getURI();
    
  	  EObject modelObject = designator.getObject();
  	  if (modelObject instanceof EPackage)
  	  {
  		EPackage ePackage = (EPackage)modelObject;
        if ((ExtendedMetaData.INSTANCE.getNamespace(ePackage) == null) && 
        	(isValidRefName(designator.getRefName())))
        {
        	result = designator.getRefName();
        }
        else 
        {
          if (ePackage.eResource() != null)
    	    xsdURI = ePackage.eResource().getURI();
        
          if (xsdURI!=null)
          {
            // return a URI relative to the mapping file if possible
            if (mappingURI!=null)
            {
              URI relativeURI = xsdURI.deresolve(mappingURI, false, true, true);
              if (relativeURI != null)
         	      result = relativeURI.toString();
            }
            else
              result = xsdURI.toString();
          }
        }
  	  }
    }
    catch (Exception e)
    {
    	result = new String();
    }
    return result;
	}


	/**
	 * Returns true if the refName is set; false otherwise.
	 * @param refName
	 * @return true if the refName is set; false otherwise.
	 */
    private boolean isValidRefName(String refName) 
	{
	  boolean result = false;
	  if ((refName != null) && (refName.length()>0))
		  result = true;
	  return result;
    }


	/**
   * Generate a relative or absolute path based on the context 
   * (getParent links) for the specified designator.
   */
	protected String getNodePathForDesignator(MappingDesignator designator)
	{
    StringBuffer sb = new StringBuffer();
    List contextList = getContextList(designator);
    String firstSegment = null;
    
	for (int i = 0; i < contextList.size(); i++) 
	{
	  MappingDesignator contextDesignator = (MappingDesignator) contextList.get(i);

	  if (i == 0) 
	  {
		firstSegment = getFirstSegment(contextDesignator, designator);
		if ((firstSegment != null) && (firstSegment.length() > 0)) 
		{
		  sb.append(firstSegment);
		  if ((contextList.size() >= 2)) 
		  {
			MappingDesignator second = (MappingDesignator) contextList.get(1);
			if (second.getIsParentDelta().booleanValue()) 
			{
			  // Second designator is modified copy of
			  // 1st context designator.
	       	  // Ignore index of first, use index from second
			  sb.append(getIndex(second));
			  // Don't process second designator again
			  i++;
			} 
			else
			{
			  // Second designator is not clone
			  // it will be processed normally below.
			}
		  }
		  else
		  {
			// Context list has only one designator
			// is path absolute?
			sb.append(getIndex(contextDesignator));
		  }
		}
	  }
	  else 
	  {
		if (sb.toString().length() > 0)
			sb.append(PATH_SEPARATOR);
		// create the remaining path segments
		sb.append(getSegment(contextDesignator.getObject()) + getIndex(contextDesignator));
	  }
	}
	return sb.toString();
  }

 /**
   * Handles non-root designators.
   * If the parent is an EPackage object then we need to resolve the next 
   * path segment using the schema model. This supports incremental
   * loading/building of the corresponding ecore model. Otherwise, the path
   * segments get resolved through the ecore model exclusively.
   */
  protected EObject getModelObjectForPath(EObject eParent, String[] path, MappingDesignator designator, MappingDesignator parentDesignator)
  {
    if ((path == null) || (path.length < 1))
    {
      designator.setParent(parentDesignator);
      designator.setObject(eParent);
      designator.setIsParentDelta(new Boolean(true));
      return eParent;
    }
    
  	EObject result = null;
    if (path[0].equals(Dot))
    {
      // remove any redundant segments
  	  result = getModelObjectForPath(eParent, removeFirstSegment(path), designator, parentDesignator);
  	}
    else
    {
      String name = getNameFromPathSegment(getPath(path[0]));
      String arrayIndex = getIndex(path[0]);
      
      if ((name != null) && (name.length() > 0)) 
      {
        EObject eobject = null;
        if (pathSegmentContainsType(path[0]))
        {
       	  // named type declaration
       	  eobject = XSDEcoreUtils.findMatchingEClass(eParent, name);
        }
        else if ((path.length==1) && (pathSegmentContainsTextNode(path[0])))
        {
      	  eobject = XSDEcoreUtils.getTextNode(eParent);
        }
        else
        {
          boolean elementContent = !path[0].startsWith(ATTRIBUTE_PREFIX);
       	  List eFeatures = XSDEcoreUtils.getContent(eParent, elementContent);
       	  eobject = XSDEcoreUtils.findMatchingEFeature(eFeatures, name);
        }
        if (eobject != null)
        {
          // if we're at the end of the path, set the designator and return the EObject, otherwise continue
          // down the path
          if (path.length==1)
          {
            designator.setObject(eobject);
            designator.setParent(parentDesignator);
			if (arrayIndex != null) 
			{
			  designator.setIndex(arrayIndex);
			}
            result = eobject;
          }
          else
          {
            // create a new designator for this eobject and set it as the parent
            MappingDesignator newParent = fMappingFactory.createMappingDesignator();
            newParent.setObject(eobject);
            if (arrayIndex != null) 
            {
			  newParent.setIndex(arrayIndex);
			}
            newParent.setParent(parentDesignator);
            designator.setParent(newParent);
            result = getModelObjectForPath(eobject, removeFirstSegment(path), designator, newParent);
          }
        }
      }
  	}
  	return result;
  }
  
  public String getVariable(MappingDesignator designator, Set set, boolean alwaysSerialize)
  {
    String result = null;
    if (designator != null)
    {
      String variable = designator.getVariable();
      if ((variable != null) && (variable.length() > 0))
      {
    	result = variable;
      }
      else
      {
        if (alwaysSerialize || serializeDesignatorVariable(designator))
        {
   	      EObject modelObject = designator.getObject();
	        String baseName = XSDEcoreUtils.getDisplayName(modelObject);
	        if ((baseName == null) || (baseName.length()==0))
	        {
		        if (XSDEcoreUtils.isInput(designator))
		        {
		          baseName = DEFAULT_INPUT_VARIABLE;
		        }
		        else
		        {
		          baseName = DEFAULT_OUTPUT_VARIABLE;
		        }
	        }
  	      if (set.contains(baseName) || set.contains(VARIABLE_PREFIX + baseName))
	        {
		        int index = 1;
		        while (set.contains(baseName + index) || set.contains(VARIABLE_PREFIX + baseName + index))
		        {
		          index++;
	          }
	          result = baseName + index;
	        }
	        else
	        {
		        result = baseName;
	        }
 	      }
      }
    }
    return result;
  }
  
  public String getVariable(MappingDesignator designator, Set set)
  {
	return getVariable(designator, set, false);
  }


	/*
	 * (non-Javadoc)
	 * @see com.ibm.ccl.mapping.resolvers.IPathResolver#serializeDesignatorVariable(com.ibm.ccl.mapping.MappingDesignator)
	 */
	public boolean serializeDesignatorVariable(MappingDesignator referencedDesignator)
	{
	  boolean isRequired = false;
	  if (referencedDesignator != null)
	  {
        Mapping mapping =  getMappingContainer(referencedDesignator);
   	    if (mapping != null)
		{
		  boolean isInput = mapping.getInputs().contains(referencedDesignator);
		  if (isInput)
		  {
		    isRequired = mapping.getInputs().size() > 1;
		  }
		}
   	  }
 	  return isRequired;
	}

	
	/* (non-Javadoc)
	 * @see com.ibm.ccl.mapping.resolvers.IPathResolver#serializePathWithDesignatorVariable(com.ibm.ccl.mapping.MappingDesignator)
	 */
	public boolean serializePathWithDesignatorVariable(MappingDesignator referencingDesignator)
	{
		boolean isRequired = false;
		if (referencingDesignator != null)
		{
		  List contextList = getContextList(referencingDesignator);
		  if ((contextList != null) && (!contextList.isEmpty()))
		  {
		  	MappingDesignator referencedDesignator = (MappingDesignator)contextList.get(0);
	  		if (requiresVariable(referencedDesignator, referencingDesignator))
				  isRequired = true;
		  }
		}
    return isRequired;
  }

  private List getContextList(MappingDesignator designator)
  {
    List result = new ArrayList();
    if (designator != null)
    {
	    result.add(designator);	
	    MappingDesignator parent = designator.getParent();
	    boolean done = false;
	    while ((parent != null) && !done)
	    {
	  	  result.add(0, parent);
  	    EObject container = parent.eContainer();
   		  if(container instanceof Mapping) 
  			  done = true;
   		  else
   			  parent = parent.getParent();
      }
    }
    return result;
  }


  /**
   * Determines if a variable is needed for the referenced and referencing designators.
   * A variable is necessary if the designator is not contextually unique and is referenced.
   * 
   * 1. if the inputs or outputs list containing the referenced designator has more than 
   *    one entry.    
   * 2. when the referencing designator differs in context from that of the referenced designator
   *    (for example, the referenced designator represents an input while the referencing 
   *    designator represents a target or vice versa).
   * 3. if the referenced designator of a mapping has a child refinement designator. 
   * 
   * 4. if the referencing designator is in a mapping other that an immediate nested
   *    mapping or the same mapping as that of the referenced designator.
   *    
   * Assumption: designators on refinements do not need their variable attributes
   *             to be serialized.
   */
  public boolean requiresVariable(MappingDesignator referencedDesignator, MappingDesignator referencingDesignator)
  {
	  boolean isRequired = false;
	
    if ((referencedDesignator != null) && (referencingDesignator != null))
    {
      Mapping referencedMapping =  getMappingContainer(referencedDesignator);
      Mapping referencingMapping =  getMappingContainer(referencingDesignator);

  	  if ((referencedMapping != null) && (referencingMapping != null))
      {
  	    boolean isInput = referencedMapping.getInputs().contains(referencedDesignator);
        if (isInput ? referencedMapping.getInputs().size() > 1 : referencedMapping.getOutputs().size() > 1)
        {
  	      isRequired = true;
	      }
	      else
	      {
	  	    List nestedMappings = referencedMapping.getNested();
		      if (referencedMapping == referencingMapping)
 		      {
 			      // designator for a group or sort refinement
 			      if (!isInput)
 				      isRequired = true;
   	  	  }
	        else if (nestedMappings.contains(referencingMapping))
	        {
	     	    // ancestor in nested mapping
	          if (isInput)
	          	isRequired = !referencingMapping.getInputs().contains(referencingDesignator);
 				    else
 				    	isRequired = !referencingMapping.getOutputs().contains(referencingDesignator);
 		      }
 		      else
 		      {
 			      // ancestor not found in refinements or nested mapping
 			      isRequired = true;
 		      }
 		    }
      }
    }
    return isRequired;
  }

  
  /**
   * The first segment of a path is handled quite differently than the other segments.
   * 1. If there needs to be an explicit reference to the context designator then 
   *    return the corresponding variable (i.e., '$'variableName).
   * 2. If the relationship is implicit then:
   *      a) if the node is the context node then return '.'.
   *      b) otherwise return the empty string since relative
   *         paths start from the context node.
   * 
   * @param contextDesignator the context designator.
   * @param designator  the current designator. 
   * @param verbose - flag indicating verbose status.
   * @return the first segment of the designator path.
   */
  protected String getFirstSegment(MappingDesignator contextDesignator, MappingDesignator designator)
  {
    String firstSegment = "";	 //$NON-NLS-1$
    	
    if ((contextDesignator != null) && (designator != null))
    {
      // relative path -- ignore the context node unless it also
      // the current node
  	  EObject current = designator.getObject();
  	  EObject context = contextDesignator.getObject();
  	  if ((current != null) && (current == context))
  	  {
  	    firstSegment = Dot;
  	  }
    }
    return firstSegment;
	}

	/**
	 * Obtains array String to append to path segment if designator has array
	 * field set. Otherwise empty String is returned
	 * 
	 * @param designator
	 * @return String
	 */
	protected String getIndex(MappingDesignator designator) {
		if (designator.getIndex() != null && designator.getIndex().length() > 0) {
			return LSquareBracket + designator.getIndex() + RSquareBracket;
		}
		return "";
	}

	/**
	 * Obtains array index from path segment
	 * 
	 * @param pathSegment
	 * @return String or null
	 */
	protected String getIndex(String pathSegment) 
	{
		if (pathSegment.endsWith("]")) 
		{
			int open = pathSegment.lastIndexOf("[");
			if (open >= 1) 
			{
				return pathSegment.substring(open + 1, pathSegment.length() - 1);
			}
		}
		return null;
	}

	/**
	 * Removes array index from the path if present
	 * 
	 * @param pathSegment
	 * @return
	 */
	protected String getPath(String pathSegment) {
		if (pathSegment.endsWith(RSquareBracket)) {
			int open = pathSegment.lastIndexOf(LSquareBracket);
			if (open >= 0) {
				return pathSegment.substring(0, open);
			}
		}
		return pathSegment;
	}

  /**
   * This returns a string segment of the path based on the type
   * of model object.
   * For a named type definition, this returns "type('"typeName"')".
   * For an attribute definition this returns "@"attributeName.
   * For an element definintion this returns elementName. 
   * 
   * Note we still need to add support for predicates which can
   * appear with element model objects.
   */
  public static String getSegment(EObject eobject)
  {
    String result = null;
    if (eobject instanceof EStructuralFeature)
    {
      int kind = ExtendedMetaData.INSTANCE.getFeatureKind((EStructuralFeature)eobject);
      switch (kind)
      {
        case ExtendedMetaData.UNSPECIFIED_FEATURE:
      	  // error case
      	  break;
        case ExtendedMetaData.SIMPLE_FEATURE:
      	  return TEXT_FUNCTION_NAME + LParen + RParen;
        case ExtendedMetaData.ATTRIBUTE_FEATURE:
        case ExtendedMetaData.ATTRIBUTE_WILDCARD_FEATURE:
      	  return XSDEcorePathResolverAdapter.ATTRIBUTE_PREFIX + XSDEcoreUtils.getDisplayName((EStructuralFeature)eobject);
        case ExtendedMetaData.ELEMENT_FEATURE:
        case ExtendedMetaData.ELEMENT_WILDCARD_FEATURE:
      	  return XSDEcoreUtils.getDisplayName((EStructuralFeature)eobject);
      
        case ExtendedMetaData.GROUP_FEATURE:
      	  // error case
      	  break;
      }
    }
    else if (eobject instanceof EClass)
    {
      EClass eClass = 	(EClass)eobject;
      String name = XSDEcoreUtils.getDisplayName(eClass);
      if ((name != null) && (name.length() > 0))
        result = TYPE_FUNCTION_NAME + LParen + SingleQuote + name + SingleQuote + RParen;
    }  
    return result;
  }


  /**
   * Returns the name from the specified path segment
   */
  public static String getNameFromPathSegment(String segment)
  {
	  String result = null;
	  if (segment != null)
	  {
		  if (segment.startsWith(ATTRIBUTE_PREFIX))
		  {
			  // an attribute, remove the atsign
			  result = segment.substring(ATTRIBUTE_PREFIX.length());
		  }
		  else if (pathSegmentContainsType(segment))
		  {
			  // a type, get the name inside the function call
		    int start = segment.indexOf(LParen + SingleQuote);
		    int end = segment.indexOf(SingleQuote + RParen);
		    if ((start != -1) && ((end-start)>2))
		    {
			    result = segment.substring(start+2, end).trim();
			    if (result.length() < 1)
			    {
			  	  // empty string -- return null
			  	  result = null;
			    }
		    }
		  }
		  else if (pathSegmentContainsTextNode(segment))
		  {
			result = TEXT_FUNCTION_NAME;
		  }
		  else
		  {
			  // an element, remove predicate if present
	      int index = segment.indexOf(LSquareBracket);
	      if (index != -1)
	      {
		      result = segment.substring(0, index).trim();
	      }
	      else
	      {
	  	    result = segment.trim();
	      }
		    if (result.length() < 1)
		    {
			    // empty string -- return null
			    result = null;
		    }
		  }
	  }
	  return result;
  }


  /**
   * Removes the first segment from the path.
   */
  public static String[] removeFirstSegment(String[] array)
  {
    if (array==null || array.length<1)
      return new String[0];
    
    String[] temp = new String[array.length-1];
    System.arraycopy(array, 1, temp, 0, temp.length);
    return temp;
  }

  
  /**
   * Removes the variable if present from the path segment.
   */
  public static String[] removeVariableFromPathSegment(String[] segments)
	{
  	String[] result = segments;
    if (segments!=null && segments.length>0)
    {
  	  // remove the variable from the segments path
      if (segments[0].startsWith(VARIABLE_PREFIX))
        result = removeFirstSegment(segments);
    }
    return result;
	}

  /**
   * Removes the variable if present from the path segment.
   */
  public static String[] removeDotFromPathSegment(String[] segments)
	{
  	String[] result = segments;
    if (segments!=null && segments.length>0)
    {
  	  // remove the variable from the segments path
      if (segments[0].equals(Dot) || segments[0].startsWith(Dot + LSquareBracket))
        result = removeFirstSegment(segments);
    }
    return result;
	}
  
  /**
   * Returns the index of the predicate in the path segment.
   * 
   * Note we still need to add support for predicates which can
   * appear with element model objects.
   */
  public static int getIndexFromPathSegment(String segment)
  {
	  int result = -1;
	  if (segment != null)
	  {
	    int start = segment.indexOf(LSquareBracket);
	    int end = segment.indexOf(RSquareBracket);
	    if ((start != -1) && ((end-start)>1))
	    {
		    String resultStr = segment.substring(start+1, end).trim();
		    if (resultStr.length() > 0)
		    {
		  	  try
		  	  {
		  		  result = Integer.parseInt(resultStr);
		  	  }
		  	  catch (Exception e)
		  	  {
		  		  result = -1;
		  	  }
		    }
	    }
	  }
	  return result;
  }

  /**
   * Returns true if the specified path segment contains a text node (i.e., text()).
   */
  public static boolean pathSegmentContainsTextNode(String segment) 
  {
	if (segment != null)
	{
	  return segment.trim().equals(TEXT_FUNCTION_NAME + LParen + RParen);
    }
	return false;
  }

  /**
   * Returns true if the specified path segment contains a type definition.
   */
  public static boolean pathSegmentContainsType(String segment)
  {
	  if (segment != null)
	  {
	    return segment.trim().startsWith(TYPE_FUNCTION_NAME + LParen);
	  }
	  else
	  {
	    return false;
	  }
  }

  
  /**
   * Returns the resource set of the mapping resource.
   */
  protected ResourceSet getResourceSet()
  {
    if (fResourceSet==null)
    {
    	MappingResources mappingResources = new MappingResources();
        fResourceSet = mappingResources.getResourceSet(null);
    }
    return fResourceSet;
  }


  /**
   * Set the resource set.
   */
  protected void setResourceSet(ResourceSet rs)
  {
    fResourceSet = rs;
  }


  /**
   * Load the resource with the given path using the 
   * mapping resource as the base URI.
   */
  protected Resource loadResource(String path, Resource mappingResource)
  {
    URI uri = null;
    
    if (mappingResource==null)
    {
      uri = URI.createPlatformResourceURI(path);
    }
    else
    {
      uri = URI.createURI(path);
      uri = uri.resolve(mappingResource.getURI());
    }
    
    ResourceSet resourceSet = getResourceSet();
    resourceSet.getLoadOptions().put(MAPPING_URI, mappingResource.getURI());
    Resource resource = resourceSet.getResource(uri, true);
    try
    {
      resource.load(Collections.EMPTY_MAP);
      for (Iterator iter = resource.getContents().iterator(); iter.hasNext();)
      {
        Object element = iter.next();
       
        // handle XSD files
        if (element instanceof XSDSchema)
        {
          EPackage ePackage = fEcoreBuilder.getEPackage((XSDSchema)element);
          fEcoreBuilder.generate((XSDSchema)element);
          resource.getContents().remove(element);
          resource.getContents().add(ePackage);
          break;
        }
      }
      
    }
    catch (IOException e)
    {
      //TODO: log this exception
      e.printStackTrace();
    }
    finally
    {
        resourceSet.getLoadOptions().remove(MAPPING_URI);
    }
    return resource;
  }

  
  /**
   * Returns the EPackage object from the contents of the resource.
   * Assumes that there s only one EPackage per resource.
   */
 	private EPackage getEPackageFromResource(Resource res)
 	{
 	  EPackage result = null;
      if (res != null)
      {
        for (Iterator iter = res.getContents().iterator(); iter.hasNext();)
        {
          Object element = iter.next();       
          if (element instanceof EPackage)
          {
            result = (EPackage)element;
          }
        }
      }
      return result;
	}

 	
	/**
   * Returns the Mapping object containing the specified mapping designator.
	 */
	protected Mapping getMappingContainer(MappingDesignator mappingDesignator)
	{
		Mapping result = null;
		if (mappingDesignator != null)
		{
		  EObject o = mappingDesignator.eContainer();
		  while (o != null)
		  {
		    if (o instanceof Mapping)
		    {
		      result = (Mapping)o;
		      break;
		    }
		    else 
		    {
			    o = o.eContainer();
		    }
		  }
		}
		return result;
	}
	
  /**
   * Given a namespace return the corresponding EPackage object.
   * @param namespace a given namespace.
   * @return the corresponding EPackage object.
   */
	public EPackage getEPackage(String namespace)
	{
    if (namespace != null)
		  return (EPackage)fEcoreBuilder.getTargetNamespaceToEPackageMap().get(namespace);
    else
    	return null;
	}
}
