/*******************************************************************************
 *  Licensed Materials - Property of IBM
 * (C) Copyright IBM Corp.  2001, 2006.  All Rights Reserved.
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.ccl.mapping.xsd.resources.extension;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import com.ibm.ccl.mapping.xsd.resources.IXSDResourceProvider;

/**
 * Maintains the registered XSD resource providers.
 */
public class XSDResourceProviderRegistry
{
  /**
   * A map with descriptors keyed by the file extension supported by their
   * resource factory.
   */
  private Map extensionToProviderMap;

  /**
   * The one and only registry instance.
   */
  private static XSDResourceProviderRegistry instance;

  /**
   * Private to enforce the singleton pattern.
   */
  private XSDResourceProviderRegistry()
  {
    extensionToProviderMap = new HashMap();
  }

  /**
   * Provides the one and only instance of this registry.
   * 
   * @return the only XSDResourceProviderRegistry
   */
  public static XSDResourceProviderRegistry getInstance()
  {
    if (instance == null)
    {
      instance = new XSDResourceProviderRegistry();
      XSDResourceProviderRegistryReader reader = new XSDResourceProviderRegistryReader(instance);
      reader.readRegistry();
    }

    return instance;
  }

  /**
   * Adds a resource provider to the registry.
   * 
   * @param provider the {@link IXSDResourceProvider} to add. Must not be null.
   */
  void addResourceProvider(IXSDResourceProvider resourceProvider)
  {
    extensionToProviderMap.put(resourceProvider.getFileExtension(), resourceProvider);
  }

  /**
   * Provides the map with registered descriptors keyed by the file extension.
   * 
   * @return a non-modifiable list with the registered descriptors.
   */
  public Map getResourceProviders()
  {
    return Collections.unmodifiableMap(extensionToProviderMap);
  }

  /**
   * Determines if the registry has a provider registered for the given file
   * extension.
   * 
   * @param fileExtension the file extension to look-up
   * @return true if the registry has a descriptor registered for the given file
   *         extension, false otherwise.
   */
  public boolean hasProviderForExtension(String fileExtension)
  {
    return extensionToProviderMap.containsKey(fileExtension);
  }
}
