/*******************************************************************************
 *  Licensed Materials - Property of IBM
 * (C) Copyright IBM Corp.  2001, 2006.  All Rights Reserved.
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.ccl.mapping.xsd.util;

import java.io.File;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IResource;
import org.eclipse.emf.ecore.EAnnotation;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EDataType;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.EcorePackage;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.util.ExtendedMetaData;
import org.eclipse.xsd.XSDAttributeDeclaration;
import org.eclipse.xsd.XSDComplexTypeDefinition;
import org.eclipse.xsd.XSDElementDeclaration;
import org.eclipse.xsd.XSDNamedComponent;
import org.eclipse.xsd.XSDSchema;
import org.w3c.dom.Document;
import org.xml.sax.InputSource;

import com.ibm.ccl.mapping.InlineRefinement;
import com.ibm.ccl.mapping.Mapping;
import com.ibm.ccl.mapping.MappingDeclaration;
import com.ibm.ccl.mapping.MappingDesignator;
import com.ibm.ccl.mapping.SemanticRefinement;
import com.ibm.ccl.mapping.SubmapRefinement;
import com.ibm.ccl.mapping.codegen.util.ModelUtils;
import com.ibm.ccl.mapping.xsd.XSDEcoreBuilder;
import com.ibm.ccl.mapping.xsd.XSDEcoreExtendedMetaData;
import com.ibm.ccl.mapping.xsd.XSDEcorePathResolverAdapter;

public class XSDEcoreUtils 
{
  protected static final String ANONYMOUS_TYPE_EXTENSION = "_._type"; //$NON-NLS-1$
  protected static final String EOBJECT_NAME = "EObject"; //$NON-NLS-1$
  protected static final String ANYTYPE_NAME = "anyType"; //$NON-NLS-1$
  
  public static boolean	isRequired(EStructuralFeature eFeature)
  {
    return eFeature.getLowerBound() == 1;
  }
  
  public static boolean hasText(EStructuralFeature feature)
  {
	boolean result = false;
	EClassifier eClassifier = feature.getEType();
	
	if (eClassifier instanceof EDataType)
	{
		return true;
	}
	else if (eClassifier instanceof EClass)
	{
	   switch (ExtendedMetaData.INSTANCE.getContentKind((EClass)eClassifier))
	   {
	      case ExtendedMetaData.SIMPLE_CONTENT:
	      case ExtendedMetaData.MIXED_CONTENT:
	      {
	         result = true;
	         break;
	      }
	   }
	}
    return result;
  }
  
  /**
   * @deprecated use isElement(EObject).
   */
  public static boolean isElement(EStructuralFeature feature)
  {
    boolean result = false;
    if (feature != null)
    {
      switch (ExtendedMetaData.INSTANCE.getFeatureKind(feature))
      {
        case ExtendedMetaData.ELEMENT_FEATURE:
        case ExtendedMetaData.ELEMENT_WILDCARD_FEATURE:
        case ExtendedMetaData.GROUP_FEATURE:
        {
          result = true;
          break;
        }
      }
    }
    return result;
  }
  	  
  public static boolean isElement(EObject eobject)
  {
    boolean result = false;
    if (eobject instanceof EStructuralFeature)
    {
      switch (XSDEcoreExtendedMetaData.INSTANCE.getFeatureKind((EStructuralFeature)eobject))
      {
        case ExtendedMetaData.ELEMENT_FEATURE:
        case ExtendedMetaData.ELEMENT_WILDCARD_FEATURE:
        case ExtendedMetaData.GROUP_FEATURE:
        {
          result = true;
          break;
        }
      }
    }
    return result;
  }

  public static boolean isType(EObject eobject)
  {
    return (eobject instanceof EClassifier); 
  }

  /**
   * Returns true if the designator represents a type.
   * @param designator a MappingDesignator object.
   * @return true if the designator represents a type.
   */
  public static boolean isType(MappingDesignator designator)
  {
	boolean result = false;
	if (designator != null)
	{
	  result = (designator.getObject() instanceof EClassifier);
	}
	return result;
  }

  /**
   * Returns the type of the specified EObject.
   * @param eobject an EObject.
   * @return the type of the specified EObject.
   */
  public static EObject getType(EObject eobject)
  {
	EObject result = null;
	if (eobject instanceof EStructuralFeature)
	{
	  result = ((EStructuralFeature)eobject).getEType();
	}
	else if (eobject instanceof EClassifier)
	{
	  result = eobject;
    }
	return result;
  }

  /**
   * Return the type infomation for the given EObject.
   * @param eobject an EObject object.
   * @return the type infomation for the given type.
   */
  public static int getTypeInfo(EObject eobject)
  {
	 int result = XSDEcoreConstants.SIMPLE;
     if (eobject instanceof EClass)
     {
       if (ExtendedMetaData.INSTANCE.getContentKind((EClass)eobject) == ExtendedMetaData.SIMPLE_CONTENT)
       {
  	     result = XSDEcoreConstants.COMPLEX_WITH_SIMPLE_CONTENT;
       }
       else
       {
          result = XSDEcoreConstants.COMPLEX_WITH_STRUCTURED_CONTENT;
      }
    }
    return result;
  }
  
  /**
   * Returns the name of the specified EObject.
   * @param eobject an EObject.
   * @return the name of the specified EObject.
   */
  public static String getTypeName(EObject eobject)
  {
	String result = null;
	if (eobject instanceof EStructuralFeature)
	{
	  result = ((EStructuralFeature)eobject).getEType().getName();
	}
	else if (eobject instanceof EClassifier)
	{
	  result = ((EClassifier)eobject).getName();
	}
	return result;
  }

  /**
   * Returns true if the specified eobject can occur multiple times within its context.
   * @param eobject an EObject.
   * @return true if the specified eobject can occur multiple times.
   */
  public static boolean isRepeatable(EObject eobject)	
  {
	boolean result = false;
	if (eobject instanceof EStructuralFeature)
	{
	  result = (((EStructuralFeature)eobject).getUpperBound() > 1 || 
		        ((EStructuralFeature)eobject).getUpperBound() == -1);
	}
	return result;
  }

  /**
   * Returns true if the specified eobject can occur multiple times within its context.
   * @param eobject an EObject.
   * @return true if the specified eobject can occur multiple times.
   */
  public static boolean isRepeatableDesignator(MappingDesignator designator)	
  {
  	boolean result = false;
  	if (designator != null)
  	{
      if (isRepeatable(designator.getObject()))
      {
  	    result = ((designator.getIndex() == null) || (designator.getIndex().length() == 0));
      }
  	}
  	return result;
  }

  /**
   * Returns true if the specified eobject is optional within its context.
   * @param eobject an EObject.
   * @return true if the specified eobject is optional.
   */
  public static boolean isOptional(EObject eobject)
  {
	boolean result = true;
	if (eobject instanceof EStructuralFeature)
	{
	  result = (((EStructuralFeature)eobject).getLowerBound()== 0) && 
	            (((EStructuralFeature)eobject).getUpperBound()== 1);
	}
  	return result;
  }

  /**
   * Returns true if the specified mapping is a element reference mapping.
   * A mapping is considered an element reference mapping:
   *    1. if the mapping is a submap refinement and
   *    2. both the inputs and outputs contain a single element
   *
   * @param mapping a Mapping object.
   * @return true if the specified mapping is a reference mapping.
   */
  public static boolean isElementMappingReference(Mapping mapping)
  {
	boolean result = false;
	if (mapping != null)
	{
	  SubmapRefinement refinement = ModelUtils.getSubmapRefinement(mapping);
	  if (refinement != null)
	  {
	    MappingDeclaration mappingDeclaration = ((SubmapRefinement)refinement).getRef();
	    if (mappingDeclaration != null)
	 	{
	      List inputs = mappingDeclaration.getInputs();
	 	  List outputs = mappingDeclaration.getOutputs();
	 	  result = ((!inputs.isEmpty()) && (!isType((MappingDesignator)inputs.get(0))) &&
	 		        (!outputs.isEmpty()) && (!isType((MappingDesignator)outputs.get(0))));
	 	}
	  }
	}
	return result;
  }

  /**
   * Returns true if the specified mapping is a element inline mapping.
   * A mapping is considered an elementinline mapping:
   *    1. if the mapping is an inline refinement and
   *    2. both the inputs and outputs contain a single element
   *
   * @param mapping a Mapping object.
   * @return true if the specified mapping is an element inline mapping.
   */
  public static boolean isElementInlineMapping(Mapping mapping)
  {
	boolean result = false;
	if (mapping != null)
	{
	  if (ModelUtils.hasInlineRefinement(mapping))
	  {
        List inputs = mapping.getInputs();
	 	List outputs = mapping.getOutputs();
	 	result = ((!inputs.isEmpty()) && (!isType((MappingDesignator)inputs.get(0))) &&
	 	          (!outputs.isEmpty()) && (!isType((MappingDesignator)outputs.get(0))));
	  }
	}
	return result;
  }

  /**
   * Returns true if the specified mapping is a type reference mapping.
   * A mapping is a considered a type reference if:
   *    1. the mapping contains a submap refinement and
   *    2. at least one of the inputs or outputs is a type. 
   *    
   * @param mapping a Mapping object.
   * @return true if the specified mapping is a type reference mapping.
   */
  public static boolean isTypeMappingReference(Mapping mapping)
  {
  	boolean result = false;
  	if (mapping != null)
  	{
 	  SubmapRefinement refinement = ModelUtils.getSubmapRefinement(mapping);
 	  if (refinement != null)
   	  {
        MappingDeclaration mappingDeclaration = ((SubmapRefinement)refinement).getRef();
  	    if (mappingDeclaration != null)
  		{
  	      boolean inputsContainType = false;
  	      for (Iterator i = mappingDeclaration.getInputs().iterator(); i.hasNext();)
  	      {
  	        if (isType((MappingDesignator)i.next())) 
	  	    {
	  	      inputsContainType = true;
	  	      break;
	  	    }
	  	  }
	  	  List outputs = mappingDeclaration.getOutputs();
	  	  result = inputsContainType || ((!outputs.isEmpty()) && isType((MappingDesignator)outputs.get(0)));
	  	}
	  }
	}
	return result;
  }

  /**
   * Returns the content of the given parent. Parent should be an EPackage, 
   * a type (EClass object) or an element (EStructuralFeature object).
   * This returns either a list of attributes or a list of elements
   * depending on the value of the specified flag.
   * @param parent a type or an element
   * @param elementContent flag indicating type of content requested
   * @return the content of the given parent.
   */
  public static  List getContent(EObject parent, boolean elementContent)
	{
  	List result = Collections.EMPTY_LIST;
  	EClass eClass = null;
    	
  	if (parent instanceof EClass)
  	{
   		eClass = (EClass)parent;
   	}
   	else if (parent instanceof EStructuralFeature)
   	{
   		EClassifier eClassifier = ((EStructuralFeature)parent).getEType();
   		if (eClassifier instanceof EClass)
   			eClass = (EClass)eClassifier;
   	}	
   	else if (parent instanceof EPackage)
   	{
   		eClass = ExtendedMetaData.INSTANCE.getDocumentRoot((EPackage)parent);
   	}
   	if (eClass != null)
   	{
      // get the child attribute or element depending on which is specified in the path
      if (elementContent)
       	result = getAllElements(eClass);
      else
        result = getAllAttributes(eClass);
    }
    return result;
  }
	
  public static boolean hasNamedComplexType(EObject eobject)
  {
	boolean result = false;
	if (eobject instanceof EStructuralFeature)
	{
		result = isNamedComplexType(((EStructuralFeature)eobject).getEType());
	}
	return result;
  }

  public static boolean isNamedComplexType(EClassifier eClassifier)
  {
	return (eClassifier instanceof EClass) && (!isAnonymous(eClassifier));
  }

  public static boolean isGlobalElement(EObject eobject)
  {
	boolean result = false;
	if (eobject instanceof EStructuralFeature)
	{
	  EStructuralFeature eFeature = ((EStructuralFeature) eobject);
	  if (isElement(eFeature))
  	    result = (getCorrespondingGlobalElement(eFeature)!= null) && hasComplexType(eFeature);
	}
	return result;
  }
	
  public static boolean hasComplexType(EStructuralFeature eFeature)
  {
	return (eFeature.getEType() instanceof EClass);
  }

  public static boolean isAnonymous(EClassifier eclassifier)
  {
	String name = ExtendedMetaData.INSTANCE.getName(eclassifier);
	return (name != null) && (name.endsWith(ANONYMOUS_TYPE_EXTENSION));
  }

  /**
   * Returns the EStructuralFeature object with the specified named.
   */
  public static EStructuralFeature findMatchingEFeature(List eFeatures, String featureName)
  {
    EStructuralFeature 	 result = null;
    if ((eFeatures != null) && (featureName != null) && (featureName.length()>0))
    {
      for (int i=0; i<eFeatures.size(); i++)
      {
	    EStructuralFeature eFeature = (EStructuralFeature)eFeatures.get(i);
	    if (featureName.equals(getDisplayName(eFeature)))
	    {
	    	result = eFeature;
        break;
	  }
    }
  }
  return result;
  }
	

  public static EStructuralFeature getCorrespondingGlobalElement(EStructuralFeature feature)
  {
	EStructuralFeature result = null;
	if (feature != null)
	{
	  String namespace = null;
	
	  // workaround bug in BasicExtendedMetaData
	  EAnnotation eAnnotation = feature.getEAnnotation(ExtendedMetaData.ANNOTATION_URI);
      if (eAnnotation != null)
      {
	    namespace = (String)eAnnotation.getDetails().get("namespace"); //$NON-NLS-1$
        if ("##targetNamespace".equals(namespace)) //$NON-NLS-1$
	    {
           EPackage ePackage = feature.getEContainingClass().getEPackage();
          
           if (ePackage != null)
           {
        	 EClass eClass = ExtendedMetaData.INSTANCE.getDocumentRoot(ePackage);
        	 List features = ExtendedMetaData.INSTANCE.getAllElements(eClass);
        	 String name = getDisplayName(feature);
        	 for (Iterator i=features.iterator(); i.hasNext();)
        	 {
        	   EStructuralFeature candidate = (EStructuralFeature)i.next();
        	   if ((name.equals(getDisplayName(candidate))) &&
              	   (feature.getEType() == candidate.getEType()))
        	   {
        	   	 result = candidate;
        	   	 break;
        	   }
        	 }
           }
        }
        else if (namespace != null)
        {
          result = ExtendedMetaData.INSTANCE.getElement(namespace, getDisplayName(feature));	
        }
        	
      }
	  if (result == null)
      {
    	result = feature;
      }
    }
    return result;
  }

  public static String getDisplayName(EObject eObject)
  {
    String result = null;
    if (eObject instanceof EClassifier)
    {
      EClassifier classifier = (EClassifier)eObject;
      if (isFeatureMap(classifier))
      {
        return "";
      }
      else if (isAnyType(classifier))
      {
        // Simply return anyType
        result = ANYTYPE_NAME;
      }

      else
      {
        result =  ExtendedMetaData.INSTANCE.getName(classifier);
      }
      
      if ((result == null) || (result.endsWith("_._type")))
      {
        result = classifier.getName(); //$NON-NLS-1$
      }
    }
    else if (eObject instanceof EStructuralFeature)
    {
      EStructuralFeature feature = (EStructuralFeature)eObject;
      
      if (isWildcard(feature))
      {
        if (isMixedContent(feature))
        {
          result = XSDEcorePathResolverAdapter.MIXED_NAME;
        }
        else
        {
       	  // Simply return any and anyAttribute
          result = feature.getName();
        }
      }
      else
      {
        if (isSimpleContent(feature))
          result = XSDEcorePathResolverAdapter.TEXT_NODE_NAME;
        else
    	  result =  ExtendedMetaData.INSTANCE.getName(feature);
      }
      if (result == null) ((EStructuralFeature)eObject).getName();
    }
    return result;
  }
 
  protected static String getDisplayName(MappingDesignator designator)
  {
  	String result = null;
  	if (designator != null)
  	{
  		result = getDisplayName(designator.getObject());
  	}
  	return result;
  }

	public static List getAttributes(EClass eClass)
	{
	  return  ExtendedMetaData.INSTANCE.getAttributes(eClass);
	}

	public static List getAllAttributes(EClass eClass)
	{
	  return  ExtendedMetaData.INSTANCE.getAllAttributes(eClass);
	}

	public static List getElements(EClass eClass)
	{
	  return ExtendedMetaData.INSTANCE.getElements(eClass);
	}

	public static List getAllElements(EClass eClass)
	{
	  return ExtendedMetaData.INSTANCE.getAllElements(eClass);
	}

	public static EClass findMatchingEClass(EObject parent, String name) 
	{
	  EClass result = null; 
	  EClassifier eClassifier = null;	
	  if (parent instanceof EPackage)
	    eClassifier = ExtendedMetaData.INSTANCE.getType((EPackage)parent, name);
	  else if (parent instanceof EStructuralFeature)
		eClassifier = ((EStructuralFeature)parent).getEType();

       if (eClassifier instanceof EClass)
		 result = (EClass)eClassifier;

	  return result;
	}
	
  public static String getDefaultMappingDeclarationName(MappingDeclaration mappingDeclaration, List unavailableNames)
  {
	String result = null;
	String basename = properName(getTargetName(mappingDeclaration));;
	String sourceName = properName(getSourceName(mappingDeclaration));
	if ((basename != null) && (basename.length() != 0))
	{
	  if ((sourceName != null) || (basename.length() != 0))
		 basename = sourceName + "To" + basename; //$NON-NLS-1$

	  if ((unavailableNames != null) && (unavailableNames.contains(basename)))
	  {
		int index = 2;
		while (unavailableNames.contains(basename + index))
		{
		  index++;
		}
		result = basename + index;
	  }
	  else
	  {
		result = basename;
	  }
	}
	else
	{
	  // no targets so use source name
	  result = sourceName;
	}
	return result;
  }

  public static String getSourceName(MappingDeclaration mappingDeclaration) 
  {
    StringBuffer sb = new StringBuffer();
	List designators = mappingDeclaration.getInputs();
	if ((designators != null) & (designators.size() > 0)) 
	{
	  sb.append(getDisplayName(((MappingDesignator) designators.get(0)).getObject()));
	  if (designators.size() > 1)
		sb.append(XSDEcoreConstants.AND_OTHERS);
	}
	return sb.toString();
  }

  public static String getTargetName(MappingDeclaration mappingDeclaration) 
  {
	StringBuffer sb = new StringBuffer();
	for (Iterator i = mappingDeclaration.getOutputs().iterator(); i.hasNext();) 
	{
	  MappingDesignator designator = (MappingDesignator) i.next();
	  sb.append(properName(getDisplayName(designator.getObject())));
	}
	return sb.toString();
  }

  public static String properName(String name) 
  {
	StringBuffer sb = new StringBuffer();
	if ((name != null) && (name.length() > 0)) 
	{
	  sb.append(Character.toUpperCase(name.charAt(0)));
	  sb.append(name.substring(1));
	  return sb.toString();
	}
	return name;
  }
  
 
  public static MappingDesignator getMatchingRootDesignator(List designators, Resource resource) 
  {
	MappingDesignator result = null;
	try
	{
      String uri = resource.getURI().toString();
      for (Iterator i = designators.iterator(); i.hasNext();)
	  {
	    MappingDesignator designator = (MappingDesignator)i.next();
	    if (designator.getObject() instanceof EPackage)
	    {
	      String nextUri = getURI((EPackage)designator.getObject());
	      if (uri.equals(nextUri))
	      {
	        result = designator;
	      }
	    }
	  }
	}
	catch (Exception e)
	{
	  result = null;
	}
	return result;
  }

  public static EObject getModelObject(XSDNamedComponent component, EPackage ePackage) 
  {
	EObject result = null;
	try
	{
 	  if (component instanceof XSDComplexTypeDefinition)
	  {
		result = findMatchingEClass(ePackage, component.getName());
	  }
	  else if (component instanceof XSDAttributeDeclaration)
	  {
	    List eFeatures = getContent(ePackage, false);
		result = findMatchingEFeature(eFeatures, component.getName());
	  }
	  else if (component instanceof XSDElementDeclaration)
	  {
		List eFeatures = getContent(ePackage, true);
		result = findMatchingEFeature(eFeatures, component.getName());
	  }
	}
	catch (Exception e)
	{
	  result = null;
	}
	return result;
  }


  public static String getURI(EPackage ePackage) 
  {
    String result = null;
    try
    {
      if ((ePackage != null) && (ePackage.eResource() != null))
      {
    	result = ePackage.eResource().getURI().toString();
      }
    }
    catch (Exception e)
    {
      result = null;
    }
    return result;
  }  

  public static EPackage getEPackage(XSDSchema schema) 
  {
	EPackage result = null;
	try
	{
	  Resource resource = schema.eResource();
	  XSDEcoreBuilder fEcoreBuilder = new XSDEcoreBuilder();
	  fEcoreBuilder.generate(schema);
	  result = fEcoreBuilder.getEPackage(schema);
	  resource.getContents().remove(schema);
	  resource.getContents().add(result);
	}
	catch (Exception e)
	{
	  result = null;
	}
	return result;
  }

  /** 
   * We have to have our version here because of DOM incompatabilites
   * with the one in com.ibm.etools.xsl.debug plugin
   */
  public static Object getDocument(IResource resource) 
  {
	Document result = null;
	if (resource instanceof IFile)
	{
	  try
	  {
	    // Create the DOM from the input file
	    String file = ((IFile)resource).getLocation().toOSString();
	    file = new File(file).toURL().toString();

	    // Make sure we handle namespace in the XML
	    DocumentBuilderFactory dfactory = DocumentBuilderFactory.newInstance();
	    dfactory.setNamespaceAware(true);
	    DocumentBuilder docBuilder = dfactory.newDocumentBuilder();
	    Document dom = docBuilder.parse(new InputSource(file));
	    result = dom;
	  }
	  catch (Exception e)
	  {
	    result = null;
	  }
    }
	return result;
  }
	
  public static boolean isInput(MappingDesignator designator) 
  {
	boolean result = true;
	if (designator != null)
	{
	  EObject container = designator.eContainer();
	  if (container instanceof Mapping) 
	  {
		if (((Mapping) container).getOutputs().contains(designator)) 
		{
		  result = false;
		}
	  }
	}
	return result;
  }

  /**
   * Determines if the given structural feature is a FeatureMap.
   * @param feature the structural feature to check.
   * @return true if the structural feature is a feature map, false otherwise.
   */
  public static boolean isFeatureMap(EStructuralFeature feature)
  {
    if (feature == null)
    {
      return false;
    }

    EClassifier type = feature.getEType();

    return isFeatureMap(type);
  }

  public static boolean isFeatureMap(EClassifier type)
  {
    if (type != null && type.equals(EcorePackage.eINSTANCE.getEFeatureMapEntry()))
    {
      return true;
    }
    return false;
  }
  
  /**
   * Determines if the given structural feature maps to a wildcard element or
   * attribute.
   * @param feature the structural feature to test.
   * @return true if the given structural feature maps to a wildcard element or
   * attribute, false otherwise.
   */
  public static boolean isWildcard(EStructuralFeature feature)
  {
    if (feature == null)
    {
      return false;
    }

    int featureKind = ExtendedMetaData.INSTANCE.getFeatureKind(feature);
    
    boolean isWildcard = (featureKind == ExtendedMetaData.ATTRIBUTE_WILDCARD_FEATURE ||
            featureKind == ExtendedMetaData.ELEMENT_WILDCARD_FEATURE );
    
    return isWildcard;
  }

  public static boolean isWildcard(EObject eobject) 
  {
	if (eobject instanceof EStructuralFeature)
	  return XSDEcoreUtils.isWildcard((EStructuralFeature)eobject);
	else
	  return false;
  }

  /**
   * Determines if the given structural feature maps to a wildcard attribute.
   * @param feature the structural feature to test.
   * @return true if the given structural feature maps to a wildcard attribute, false otherwise.
   */
  public static boolean isAttributeWildcard(EStructuralFeature feature)
  {
    if (feature == null)
    {
      return false;
    }

    int featureKind = ExtendedMetaData.INSTANCE.getFeatureKind(feature);
    
    boolean isAttributeWildcard = (featureKind == ExtendedMetaData.ATTRIBUTE_WILDCARD_FEATURE);
    
    return isAttributeWildcard;
  }

  private static boolean isAnyType(EClassifier classifier) 
  {
    if (!(classifier instanceof EClass))
    {
      return false;
    }
    return (EOBJECT_NAME.equals(classifier.getName()));
  }
  
  public static boolean isHeadToHeadMapping(Mapping parent, Mapping child)
  {
	if ((parent == null) && (child == null))
		return false;
		
	try
	{
	  for (Iterator i = child.getInputs().iterator(); i.hasNext();)
	  {
		  MappingDesignator input = (MappingDesignator)i.next();
		  EObject eobject = input.getObject();
		  if ((eobject == null) || (eobject != input.getParent().getObject()))
		  	return false;
	  }

	  for (Iterator i = child.getOutputs().iterator(); i.hasNext();)
	  {
		  MappingDesignator output = (MappingDesignator)i.next();
		  EObject eobject = output.getObject();
		  if ((eobject == null) || (eobject != output.getParent().getObject()))
		  	return false;
	  }
	}
	catch (Exception e)
	{
		return false;
	}
	return true;
  }
  
  /**
   * Provides the list with a wildcard's namespaces.
   * 
   * @param eStructuralFeature the structural feature to test. This is assumed to represent a
   * wildcard from the source schema.
   * @return a List with the wildcard's namespaces.
   */
  public static List getWildcards(EStructuralFeature eStructuralFeature)
  {
    return (((XSDEcoreExtendedMetaData)XSDEcoreExtendedMetaData.INSTANCE).getSourceWildcards(eStructuralFeature));
  }

  public static boolean isSimpleContent(EStructuralFeature feature) 
  {
	try
	{
      return (ExtendedMetaData.INSTANCE.getFeatureKind(feature) == ExtendedMetaData.SIMPLE_FEATURE);
    }
    catch (Exception e)
	{
	  return false;
	}
  }

  public static boolean isMixedContent(EStructuralFeature feature) 
  {
	try
	{
      return ((ExtendedMetaData.INSTANCE.getFeatureKind(feature) == ExtendedMetaData.ELEMENT_WILDCARD_FEATURE) &&
    		  (feature.getName().equals("mixed")));
    }
    catch (Exception e)
	{
	  return false;
	}
  }

  public static EStructuralFeature getTextNode(EObject eObject) 
  {
	EStructuralFeature result = null;
	if (eObject instanceof EStructuralFeature)
	{
	  EClassifier eClassifier = ((EStructuralFeature)eObject).getEType();
	  if (eClassifier instanceof EClass)
	  {

		result =  XSDEcoreExtendedMetaData.INSTANCE.getSimpleFeature((EClass)eClassifier);
	  }
	}
	else if (eObject instanceof EClass)
	{
	  result =  XSDEcoreExtendedMetaData.INSTANCE.getSimpleFeature((EClass)eObject);
	}
	return result;
  }
  
  public static String getNamespace(Object model) 
  {
    String result = null;
    if (model instanceof EStructuralFeature)
    {
	  result = XSDEcoreExtendedMetaData.INSTANCE.getNamespace((EStructuralFeature) model);
	  if(result == null)
	  {
	    result = getNamespace(((EStructuralFeature) model).getEContainingClass());
	  }
    }
    else if (model instanceof EClassifier)
    {
	  result = XSDEcoreExtendedMetaData.INSTANCE.getNamespace((EClassifier) model);
    }
	return result;
  }

  public static EObject getSimpleFeature(EObject type) 
  {
	return  XSDEcoreExtendedMetaData.INSTANCE.getSimpleFeature((EClass) type);
  }

  public static EObject getMixedFeature(EObject type)
  {
	return  XSDEcoreExtendedMetaData.INSTANCE.getMixedFeature((EClass) type);
  }

  public static boolean isReservedNamespace(String namespace) 
  {
	boolean result = false;
	if (namespace != null)
	{
	  if (namespace.equals(XSDEcoreConstants.NS_URI_XML) || 
	      namespace.equals(XSDEcoreConstants.NS_URI_XMLNS) || 
	      namespace.equals(XSDEcoreConstants.NS_URI_XSI))
		result = true;
	}
	return result;
  }
  
  /**
   * Returns true if the specified mapping is an inline refinement.
   * @param mapping a Mapping object.
   * @return true if the specified mapping is an inline refinement.
   */
  public static boolean isExclusiveInlineRefinement(Mapping targetMapping) 
  {
    boolean result = ModelUtils.hasInlineRefinement(targetMapping);
	if (result)
	{
	  EObject container = targetMapping.eContainer();
	  if (container instanceof Mapping)
	  {
		Mapping baseMapping = (Mapping)container;
        MappingDesignator targetDesignator = ModelUtils.getPrimaryTargetDesignator(targetMapping);
		MappingDesignator baseDesignator = ModelUtils.getPrimaryTargetDesignator(baseMapping);
	    List targetContext = ModelUtils.getMappingContextList(baseDesignator, targetDesignator);
	    int targetContextSize = targetContext.size();
	    if (targetContextSize > 1)
	    {
	      List siblings = baseMapping.getNested();

	  	  for (Iterator i = siblings.iterator(); i.hasNext();)
	  	  {
		    Object o = i.next();
		    if ((o instanceof Mapping) && (o != targetMapping))
		    {
		      Mapping candidateMapping = (Mapping)o;
		      MappingDesignator candidateDesignator = ModelUtils.getPrimaryTargetDesignator(candidateMapping);
		      List candidateContext = ModelUtils.getMappingContextList(baseDesignator, candidateDesignator);
		      if (candidateContext.size() >= targetContextSize)
		      {
		        boolean match = true;
		      
		        for (int j = 0; j < targetContextSize; j++)
		        {
		    	  if (targetContext.get(j) != (candidateContext.get(j)))   
		    	  {
		    	    match = false;
		    	    break;
		    	  }
		    	}
		        if (match)
		        {
		          result = false;
		          break;
		        }
		      }
		    }
	  	  }
	  	}
	  }
	}
	return result;
  }
}